/*
Copyright (c) 2018, Cryst.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, 
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, 
      this list of conditions and the following disclaimer in the documentation 
      and/or other materials provided with the distribution.
    * Neither the name of the Cryst nor the names of its contributors may be 
      used to endorse or promote products derived from this software without 
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
OF THE POSSIBILITY OF SUCH DAMAGE.

Copyright (c)  2017, Cryst.
All rights reserved.

  ソースコード形式かバイナリ形式か、変更するかしないかを問わず、以下の条件を
  満たす場合に限り、再頒布および使用を許可します。

    * ソースコードを再頒布する場合、上記の著作権表示、本条件一覧、および 
      下記免責条項を含めること。
    * バイナリ形式で再頒布する場合、頒布物に付属のドキュメント等の資料に、
      上記の著作権表示、本条件一覧、および下記免責条項を含めること。
    * 書面による特別の許可なしに、本ソフトウェアから派生した製品の宣伝または
      販売促進に、Cryst の名前またはコントリビューターの名前を使用してはならない。

  本ソフトウェアは、著作権者およびコントリビューターによって「現状のまま」提供さ
  れており、明示黙示を問わず、商業的な使用可能性、および特定の目的に対する適合性
  に関する暗黙の保証も含め、またそれに限定されない、いかなる保証もありません。
  著作権者もコントリビューターも、事由のいかんを問わず、損害発生の原因いかんを問
  わず、かつ責任の根拠が契約であるか厳格責任であるか（過失その他の）不法行為であ
  るかを問わず、仮にそのような損害が発生する可能性を知らされていたとしても、
  本ソフトウェアの使用によって発生した（代替品または代用サービスの調達、使用の
  喪失、データの喪失、利益の喪失、業務の中断も含め、またそれに限定されない）
  直接損害、間接損害、偶発的な損害、特別損害、懲罰的損害、または結果損害について、
  一切責任を負わないものとします。
*/
////////////////////////////////////////////////////////////////////////////////
// コンパイル方法の例：
// gcc -O3 -Wall -W -o complex_array_info complex_array_info.c
////////////////////////////////////////////////////////////////////////////////
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <sys/stat.h>

#define IFNAMEDEFAULT  "dft_complex.dat"

typedef struct COMPLEX  { double r, i; } COMPLEX;
#pragma pack(push,2)
typedef struct HEADERCX { // COMPLEX => bfType==0x585043 (0x43,0x50,0x58)
    unsigned int   cxType, zero;    // identification 0x43='C',0x50='P',0x58='X'
    unsigned int   cxSize;          // ?? cxOffByts+cxSizeData+sizeof(message)
    unsigned int   cxDataByte;      // 16 sizeof(COMPLEX)
    unsigned int   cxOffByts;       // ?? 32+α sizeof(HEADER) +sizeof(message)
    unsigned int   cxSizeData;      // ?? lx*ly*sizeof(COMPLEX)
    signed   int   cxWidth;         // ?? lx
    signed   int   cxHeight;        // ?? ly
    ///////////////////////////////////////////////////////////////////////////
    char           message[];
} HEADERCX ;
#pragma pack(pop)
#define COMPLEXID     0x585043


void *dataFile2memory( char *fname ){
// ディスク上のファイルを、メモリーに一括で読み込む
struct stat fileInfo ;
FILE   *fp ;
char   *raw=NULL ;

    if ( (fp=fopen(fname,"rb")) == NULL ) {
        fprintf( stderr, "'%s' open error.\n", fname );               return raw;
    } else if ( stat(fname, &fileInfo) != 0 ) {
        fprintf( stderr, "'%s' stat error.\n", fname );  fclose(fp);  return raw;
    } else if( (raw=(char *)malloc(fileInfo.st_size+32)) == NULL ) {
        fprintf( stderr, "Couldn't allocate memory.\n" ) ; fclose(fp);return raw;
    } else if ( fread(raw, fileInfo.st_size, 1, fp) != 1 ) {
        fprintf( stderr, "'%s' read error.\n", fname );
        free(raw); fclose(fp);  return NULL;
    } else fclose(fp) ;
    return raw;
}

void argumentAnalysis( int argc, char *argv[], char **fname ){
int  i, st=1;   // プログラム起動時の引数解析処理
    for( i=1; i<argc; i++ ) {
        if( argv[i][0]=='-' ) {
            switch(argv[i][1]) {
                case 'h': st=0;
                default : goto GOOD_BY;
            }
        }
        else if( *fname==NULL ) *fname = argv[i];
        else goto GOOD_BY;
    }
    return;
GOOD_BY:;
    char *p=argv[0], *op="[-h] ";
    for( i=strlen(p); i>0 && p[i]!='\\' && p[i]!='/'; i-- );; if(i) i++;
    fprintf( stderr, "\nusage: %s %s[complex.dat]\n\n", &p[i], op );
    fprintf( stderr, "\t-h: This option shows only this message.\n" );
    exit(st);
}

int main( int argc, char *argv[] ){
// dft_2D がフーリエ変換した結果を格納した２次元配列情報を表示する
int   st=1;
char  *raw=NULL, *fname=NULL;

    argumentAnalysis( argc, argv, &fname );    // 引数解析
    if( fname==NULL ) fname = IFNAMEDEFAULT;   // 入力ファイル名の設定

    // データファイルの確認
    struct stat  fileInfo ;
    if ( (st=stat( fname, &fileInfo )) ) {
        fprintf(stderr, "'%s' stat error.\n", fname );  goto GOOD_BY;
    } else if( fileInfo.st_size<(int)sizeof(HEADERCX) ) {
        fprintf(stderr, "'%s' size error.\n", fname );  goto GOOD_BY;
    }
    // データの読み込み（ディスクからメモリーへ）
    if( (raw=(char*)dataFile2memory(fname))==NULL ) {
        fprintf(stderr, "'%s' read error.\n", fname );  goto GOOD_BY;
    }

    // データフォーマットの確認
    HEADERCX *d=(HEADERCX*)raw;
    if( d->cxType!=COMPLEXID  ) {
        fprintf(stderr, "'%s' unknown file\n", fname );
    } else {
        fprintf( stdout, "identifier: '%s'\n", (char*)&d->cxType );
        fprintf( stdout, "cxType: 0x%4.4X\n", d->cxType );
        fprintf( stdout, "zero:       %d\n", d->zero );
        fprintf( stdout, "cxSize:     %d\n", d->cxSize );
        fprintf( stdout, "cxDataByte: %d\n", d->cxDataByte );
        fprintf( stdout, "cxOffByts:  %d\n", d->cxOffByts );
        fprintf( stdout, "cxSizeData: %d\n", d->cxSizeData );
        fprintf( stdout, "cxWidth:    %d\n", d->cxWidth );
        fprintf( stdout, "cxHeight:   %d\n", d->cxHeight );
    }

GOOD_BY:; // 後始末とプログラムの終了
    free(raw);
    return st;
}
