/*
Copyright (c) 2018, Cryst.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, 
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, 
      this list of conditions and the following disclaimer in the documentation 
      and/or other materials provided with the distribution.
    * Neither the name of the Cryst nor the names of its contributors may be 
      used to endorse or promote products derived from this software without 
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
OF THE POSSIBILITY OF SUCH DAMAGE.

Copyright (c)  2017, Cryst.
All rights reserved.

  ソースコード形式かバイナリ形式か、変更するかしないかを問わず、以下の条件を
  満たす場合に限り、再頒布および使用を許可します。

    * ソースコードを再頒布する場合、上記の著作権表示、本条件一覧、および 
      下記免責条項を含めること。
    * バイナリ形式で再頒布する場合、頒布物に付属のドキュメント等の資料に、
      上記の著作権表示、本条件一覧、および下記免責条項を含めること。
    * 書面による特別の許可なしに、本ソフトウェアから派生した製品の宣伝または
      販売促進に、Cryst の名前またはコントリビューターの名前を使用してはならない。

  本ソフトウェアは、著作権者およびコントリビューターによって「現状のまま」提供さ
  れており、明示黙示を問わず、商業的な使用可能性、および特定の目的に対する適合性
  に関する暗黙の保証も含め、またそれに限定されない、いかなる保証もありません。
  著作権者もコントリビューターも、事由のいかんを問わず、損害発生の原因いかんを問
  わず、かつ責任の根拠が契約であるか厳格責任であるか（過失その他の）不法行為であ
  るかを問わず、仮にそのような損害が発生する可能性を知らされていたとしても、
  本ソフトウェアの使用によって発生した（代替品または代用サービスの調達、使用の
  喪失、データの喪失、利益の喪失、業務の中断も含め、またそれに限定されない）
  直接損害、間接損害、偶発的な損害、特別損害、懲罰的損害、または結果損害について、
  一切責任を負わないものとします。
*/
////////////////////////////////////////////////////////////////////////////////
// コンパイル方法の例：
// gcc -O3 -Wall -W -o dft_2D dft_2D.c
////////////////////////////////////////////////////////////////////////////////
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define IMAGEPROCESSING              dft_2D
#define FINALFNAMEIS(foname)         #foname
#define EXTRACTING(foname)           FINALFNAMEIS(foname.bmp)
#define FNAMEIS(foname)              EXTRACTING(foname)
#define IFNAMEDEFAULT                "sample.bmp"

#define saveArray2bmp32(name,a,w,h)     saveDataAsBMP32((name),(&(a)[0][0]),(w),(h))
#define saveArray2COMPLEX(name,m,a,w,h) saveDataAsCOMPLEX((name),(m),(&(a)[0][0]),(w),(h))

#ifndef M_PI
#define M_PI            3.14159265358979323846
#endif

extern int  saveDataAsBMP32( char *, void *, int, int );
extern void **alloc2D( void *, int, int, int );
extern void **alloc2Darray( int, int, int );

#pragma pack(push,1)
typedef struct PIXEL   { unsigned char B, G, R, A; } PIXEL ;
#pragma pack(pop)

typedef struct COMPLEX { double r, i; } COMPLEX;
#define AddCOMPLEX(a,b)   (COMPLEX){ (a).r + (b).r, (a).i + (b).i }
#define SubCOMPLEX(a,b)   (COMPLEX){ (a).r - (b).r, (a).i - (b).i }
#define MulCOMPLEX(a,b)   (COMPLEX){ (a).r * (b).r - (a).i * (b).i, (a).r * (b).i + (a).i * (b).r }
#define DivCOMPLEX(a,b)   (COMPLEX){ ((a).r*(b).r + (a).i*(b).i)/((a).r*(a).r + (b).i*(b).i),\
                                     ((a).i*(b).r - (a).r*(b).i)/((a).r*(a).r + (b).i*(b).i) }
#define ImeMulCOMPLEX(a,b)(COMPLEX){ (a).r * (b), (a).i * (b) }
#define ImeDivCOMPLEX(a,b)(COMPLEX){ (a).r / (b), (a).i / (b) }
#define RevCOMPLEX(a)     (COMPLEX){ (a).r ,-((a).i) }
#define RotorCOMPLEX(a)   (COMPLEX){ cos(a), sin(a) }
#define ImeDivCX          ImeDivCOMPLEX

#pragma pack(push,2)
typedef struct HEADERCX { // COMPLEX => bfType==0x585043 (0x43,0x50,0x58)
    unsigned int   cxType, zero;    // identification 0x43='C',0x50='P',0x58='X'
    unsigned int   cxSize;          // ?? cxOffByts+cxSizeData+sizeof(message)
    unsigned int   cxDataByte;      // 16 sizeof(COMPLEX)
    unsigned int   cxOffByts;       // ?? 32+α sizeof(HEADER) +sizeof(message)
    unsigned int   cxSizeData;      // ?? lx*ly*sizeof(COMPLEX)
    signed   int   cxWidth;         // ?? lx
    signed   int   cxHeight;        // ?? ly
    ///////////////////////////////////////////////////////////////////////////
    char           message[];
} HEADERCX ;
#pragma pack(pop)
#define COMPLEXID     0x585043

enum DFT_DirectionOfConversion {
     DFT,
     IDFT
}    DFT_DirectionOfConversion_END;

////////////////////////////////////////////////////////////////////////////////
void creatingTheRotorTable( COMPLEX **t, int n, double u ) {
int  i, j;
    for(i=0;i<n;i++)for(j=i;j<n;j++) t[i][j]=t[j][i]=(COMPLEX){cos(u*i*j),sin(u*i*j)};
}

void dft_TwoDimensionalWithTable( COMPLEX **ss, int n, COMPLEX **d, int lx, int ly, int flag ) {
int     x, y, u, v, elements;
double  unit;
COMPLEX t[n];

    elements=lx; unit = (flag==IDFT) ? 2*M_PI/elements : -2*M_PI/elements;
    for( creatingTheRotorTable(ss,elements,unit), y=0; y<ly; y++ ) {
        for( u=0; u<elements; u++ ) t[u]=(COMPLEX){0,0};
        for( u=0; u<elements; u++ ) {
            for( v=0; v<elements; v++ ) t[u] = AddCOMPLEX( t[u], MulCOMPLEX(d[y][v],ss[u][v]) );
        }
        if(flag==IDFT) for( u=0; u<elements; u++ ) d[y][u] = ImeDivCOMPLEX(t[u],elements);
        else           for( u=0; u<elements; u++ ) d[y][u] = t[u];
    }//////////////////////////////////////////////////////////////////////////
    elements = ly;  unit = (flag==IDFT) ? 2*M_PI/elements : -2*M_PI/elements;
    for( creatingTheRotorTable(ss,elements,unit), x=0; x<lx; x++ ) {
        for( u=0; u<elements; u++ ) t[u]=(COMPLEX){0,0};
        for( u=0; u<elements; u++ ) {
            for( v=0; v<elements; v++ ) t[u] = AddCOMPLEX( t[u], MulCOMPLEX(d[v][x],ss[u][v]) );
        }
        if(flag==IDFT) for( u=0; u<elements; u++ ) d[u][x] = ImeDivCOMPLEX(t[u],elements);
        else           for( u=0; u<elements; u++ ) d[u][x] = t[u];
    }//////////////////////////////////////////////////////////////////////////
}

void dft_TwoDimensionalStandard( COMPLEX **d, int lx, int ly, int flag ) {
int     x, y, u, v, elements;
double  unit;
COMPLEX t[(lx>ly?lx:ly)];

    elements = lx;  unit = (flag==IDFT) ? 2*M_PI/elements : -2*M_PI/elements;
    for( y=0; y<ly; y++ ) {
        for( u=0; u<elements; u++ ) t[u]=(COMPLEX){0,0}; // 
        for( u=0; u<elements; u++ ) {
            for( v=0; v<elements; v++ ) t[u] = AddCOMPLEX( t[u], MulCOMPLEX(d[y][v],RotorCOMPLEX(unit*u*v)) );
        }
        if(flag==IDFT) for( u=0; u<elements; u++ ) d[y][u] = ImeDivCOMPLEX(t[u],elements);
        else           for( u=0; u<elements; u++ ) d[y][u] = t[u];
    }//////////////////////////////////////////////////
    elements = ly;  unit = (flag==IDFT) ? 2*M_PI/elements : -2*M_PI/elements;
    for( x=0; x<lx; x++ ) {
        for( u=0; u<elements; u++ ) t[u]=(COMPLEX){0,0};
        for( u=0; u<elements; u++ )  {
            for( v=0; v<elements; v++ ) t[u] = AddCOMPLEX( t[u], MulCOMPLEX(d[v][x],RotorCOMPLEX(unit*u*v)) );
        }
        if(flag==IDFT) for( u=0; u<elements; u++ ) d[u][x] = ImeDivCOMPLEX(t[u],elements);
        else           for( u=0; u<elements; u++ ) d[u][x] = t[u];
    }//////////////////////////////////////////////////
}

void dft_TwoDimensional( COMPLEX **s, int lx, int ly, int flag ) { // upgrade
int     w = lx>ly ? lx : ly;
COMPLEX **spreadsheet = NULL;

    if( (spreadsheet=(COMPLEX **)alloc2Darray(sizeof(COMPLEX),w,w)) ) {
        dft_TwoDimensionalWithTable( spreadsheet, w, s, lx, ly, flag );
        free(spreadsheet);
    } else dft_TwoDimensionalStandard( s, lx, ly, flag );
}// See also, Euler's formula & Definition of trigonometric functions (Symmetry)


int saveDataAsCOMPLEX( char *fname, char *m, COMPLEX *b, int lx, int ly ){
int message = m==NULL ? 0 :((strlen(m)>0) ? strlen(m)+1 : 0);
#define  HEADERSIZE  sizeof(HEADERCX)
HEADERCX h = { COMPLEXID, 0, HEADERSIZE +message +lx*abs(ly)*sizeof(COMPLEX),
               sizeof(COMPLEX),HEADERSIZE+message,lx*ly*sizeof(COMPLEX), lx,ly};
FILE *fp ;

    if( (fp=fopen(fname, "wb")) == NULL ) {
        fprintf( stderr, "File '%s' cannot create.\n", fname );
        return -1;
    } else if ( fwrite( &h, sizeof(HEADERCX), 1, fp ) != 1 ) {
        fclose(fp) ; return -1;
    } else if ( m!=NULL && fwrite( m, message, 1, fp ) != 1 ) {
        fclose(fp) ; return -1;
    } else if ( fwrite( b, h.cxSizeData, 1, fp ) != 1 ) {
        fclose(fp) ; return -1;
    } else fclose(fp) ;
    return 0;
}

void interchange2D( COMPLEX **d, int lx, int ly ) {
int  x, y, hlx=lx/2, hly=ly/2;

    for( y=0; y<hly; y++ ) for( x=0; x<hlx; x++ ){
        COMPLEX t;
        t = d[y][x]; d[y][x] = d[hly +y][hlx +x]; d[hly +y][hlx +x] = t;
        t = d[y][hlx +x]; d[y][hlx +x] = d[hly +y][x]; d[hly +y][x] = t;
    }
}

void PowerSpectrum( char *fname, COMPLEX **u, PIXEL **a, int lx, int ly ) {
int    x, y, gray;
double temp, min, max, div, mag;

    min=max=u[0][0].r*u[0][0].r + u[0][0].i*u[0][0].i;
    for( y=0; y<ly; y++ ) for( x=0; x<lx; x++ ) {
        if( max<(temp=u[y][x].r*u[y][x].r + u[y][x].i*u[y][x].i) ) max = temp;
        else if( min>temp ) min = temp;
    }   min = sqrt(min);  max = sqrt(max);
    div = max>min ? log10(max-min) : 1 ;
    for( mag=255.0/div, y=0; y<ly; y++ ) for( x=0; x<lx; x++ ) {
        if( u[y][x].r==0 && u[y][x].i==0 ) a[y][x] = (PIXEL){0,0,0,0};
        else {
            gray = mag*log10(sqrt(u[y][x].r*u[y][x].r + u[y][x].i*u[y][x].i) -min);
            a[y][x] = (PIXEL){gray,gray,gray,0};
        }
    }   saveArray2bmp32( fname, a, lx, ly );
}

void complexReal2grayEx( COMPLEX **u, PIXEL **a, int lx, int ly ) {
double min, max, div;
int    hist[1024], lut[1024], **w=(int**)a, m, x, y;

    for( min=max=u[0][0].r, y=0; y<ly; y++ ) for( x=0; x<lx; x++ ) {
        if(      min>u[y][x].r ) min = u[y][x].r;
        else if( max<u[y][x].r ) max = u[y][x].r;
    }   div = max>min ? max-min : 1 ;
    if( 0<=min && max<256 ) {
        for( y=0; y<ly; y++ ) for( x=0; x<lx; x++ ) a[y][x]=(PIXEL){m=u[y][x].r,m,m,0};
    } else{
        for( x=0; x<1024; x++ ) hist[x]=lut[x]=0;
        for( y=0; y<ly; y++ ) for( x=0; x<lx; x++ ){
            hist[w[y][x]=1023.0*(u[y][x].r-min)/div]++;
        }
        int l, h, n, m=((lx*ly)/200)>0 ? ((lx*ly)/100) : 1 ;
        for(      n=l=0; (n+hist[l])<m && l<1024; l++ ) n+=hist[l];
        for( n=0,h=1023; (n+hist[h])<m && h>0;    h-- ) n+=hist[h];
    
        for( n=0; n<1024; n++ ) lut[n]=0xFFFFFF;
        for( n=0; n<=l;   n++ ) lut[n]=0;
        div = h>l ? h-l : 1 ;
        for( n=1; (l+n)<h; n++ ) { m=255*n/div; lut[l+n]= m<<16| m<<8 | m; }
        for( y=0; y<ly; y++ ) for( x=0; x<lx; x++ ) w[y][x]=lut[w[y][x]];
    }
}

void complexReal2grayBMP( char *fname, COMPLEX **u, PIXEL **a, int lx, int ly ) {
    complexReal2grayEx( u, a, lx, ly );
    saveArray2bmp32( fname, a, lx, ly );
}


int IMAGEPROCESSING( PIXEL **s, PIXEL **d, int lx, int ly ){
int  x, y, st=-1 ;
COMPLEX **w=(COMPLEX**)alloc2Darray( sizeof(COMPLEX), lx, ly );

    if( w==NULL || lx<=1 || ly<=1 ) goto GOOD_BY;
    for( y=0; y<ly; y++ ) for( x=0; x<lx; x++ ) {
        int gray = (306*s[y][x].R + 601*s[y][x].G + 117*s[y][x].B)>>10;
        w[y][x]  = (COMPLEX){gray,0};  // save gray image to real part
    }   complexReal2grayBMP( "SourceToGray.bmp", w, d, lx, ly );
    
    // フーリエ変換（濃淡画像 => 周波数表現） 四隅が低周波、中心が高周波成分
    dft_TwoDimensional( w, lx, ly, DFT );
    saveArray2COMPLEX( "dft_complex.dat", NULL, w, lx, ly );
    
    // パワースペクトラム画像作成） 中心が低周波、四隅が高周波成分
    interchange2D( w, lx, ly );  // 低周波と高周波成分の場所の入れ替え
    PowerSpectrum( "PowerSpectrum.bmp", w, d, lx, ly );
    interchange2D( w, lx, ly );  // 低周波と高周波成分の場所の入れ替え

GOOD_BY:;
    free(w);
    return st ;
}
////////////////////////////////////////////////////////////////////////////////
/// 以下、メインプログラム関連ルーチンとBMP画像を処理する共通ルーチン群
///
#include <sys/stat.h>

#define BMPID      0x4D42
#define BMPIFHSIZE 40

#pragma pack(push,2)
typedef struct HEADER {       // BMP => bfType==0x4D42 (0x42,0x4D) little endian
    unsigned short bfType;          // identification  0x42='B', 0x4D='M'
    unsigned int   bfSize;          // ?? 
    unsigned short bfReserved1;     //  0 
    unsigned short bfReserved2;     //  0 
    unsigned int   bfOffBits;       // 54  sizeof(HEADER)
    ////////////////////////////////////////////////////////////////////////////
    unsigned int   biSize;          // 40 sizeof(biSize ～ biClrImportant)
    signed int     biWidth;         // lx
    signed int     biHeight;        // ly
    unsigned short biPlanes;        //  1
    unsigned short biBitCount;      // 32
    unsigned int   biCompression;   //  0
    unsigned int   biSizeImage;     // ??
    signed int     biXPelsPerMeter; //  0
    signed int     biYPelsPerMeter; //  0
    unsigned int   biClrUsed;       //  0
    unsigned int   biClrImportant;  //  0
    ////////////////////////////////////////////////////////////////////////////
    unsigned char  lut[];
} HEADER ;
#pragma pack(pop)

void **alloc2D( void *data, int s, int lx, int ly ){
// データ（lx*ly個）の１次元配列を、array[y][x] 形式で参照できるようにする
long long w=lx, h=ly;
void **d, *base=data ;
long i;

    if( (data==NULL) || (s<1) || (lx<1) || (ly<1) ) return NULL ;
    if( (d=(void **)malloc(h*sizeof(void *))) == NULL ) return NULL ;
    for( i=0; i<ly; i++ ) d[i] = (void *)((unsigned char *)base + w*s*i) ;
    return d ;
}

#define calloc2Darray alloc2Darray
void **alloc2Darray( int s, int lx, int ly ){
// array[y][x] 形式で参照できる、横：lx、縦：ly の配列を動的に確保する
long long w=lx, h=ly;
void **d, *base ;
long i ;

    if( (s<1) || (lx<1) || (ly<1) ) return NULL ;
    if( (d=(void **)calloc(1,h*sizeof(void *)+(w*h*s))) == NULL ) return NULL ;
    base = (unsigned char *)d + h*sizeof(void *) ;
    for( i=0; i<ly; i++ ) d[i] = (void *)((unsigned char *)base + w*s*i) ;
    return d ;
}

int saveDataAsBMP32( char *fname, void *b, int lx, int ly ){
// 処理結果を 32ビット形式の BMPファイルとして書き出す。
HEADER h = {BMPID, lx*abs(ly)*sizeof(PIXEL)+sizeof(HEADER), 0, 0, sizeof(HEADER),
            BMPIFHSIZE, lx, ly, 1, 32, 0, lx*abs(ly)*sizeof(PIXEL), 0, 0, 0, 0} ;
FILE *fp ;

    if( (fp=fopen(fname, "wb")) == NULL ) {
        fprintf( stderr, "File '%s' cannot create.\n", fname );
        return -1;
    } else if ( fwrite( &h, sizeof(HEADER), 1, fp ) != 1 ) {
        fclose(fp) ; return -1;
    } else if ( fwrite( b, h.biSizeImage, 1, fp ) != 1 ) {
        fclose(fp) ; return -1;
    } else fclose(fp) ;
    return 0;
}

int bmp2pixel( void *bmp, void *pix ){
// メモリー上の BMP ファイルを、PIXEL 形式のデータに変換する
static char     f[8] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };
HEADER *h=(HEADER *)bmp;
PIXEL  *d=(PIXEL*)pix, *lut=(PIXEL*)h->lut;
int    x, y, i, top=((h->bfOffBits)-sizeof(HEADER))/sizeof(PIXEL);
int    sign, lx=h->biWidth, ly=abs((sign=h->biHeight));
int    BitCount=h->biBitCount, byteParLine=(lx*(BitCount/8)+3)&(~3);
int    byteParLineBW = (((lx+7)/8)+3)&(~3);
unsigned char *src=(unsigned char*)bmp + h->bfOffBits;
unsigned char *dst=(unsigned char*)pix, *here;

    if( BitCount == 32 ) {
        for( y=0; y<ly; y++ ) {
            here = src + byteParLine * ( sign>0 ? y :ly-1 -y );
            for( x=0; x<lx; x++ ) {
                *dst++ = *here++ ; // B
                *dst++ = *here++ ; // G
                *dst++ = *here++ ; // R
                *dst++ = *here++ ; // A
            }
        }
    } else if( BitCount == 24 ) {
        for( y=0; y<ly; y++ ) {
            here = src + byteParLine * ( sign>0 ? y :ly-1 -y );
            for( x=0; x<lx; x++ ) {
                *dst++ = *here++ ; // B
                *dst++ = *here++ ; // G
                *dst++ = *here++ ; // R
                *dst++ = 0 ;
            }
        }
    } else if( BitCount == 8 ) {
        int *u=(int*)pix, *t=(int*)h->lut;
        for( y=0, i=0; y<ly; y++ ) {
            here = src + byteParLine * ( sign>0 ? y :ly-1 -y );
            for( x=0; x<lx; x++ ) u[i++] = here[x]<top ? t[here[x]] : 0;
        }
    } else if( BitCount == 1 ) {
        #define getBW(t,x) (((t)&(f[0x7&(x)]))?1:0)
        for( y=0, i=0; y<ly; y++ ) {
            here = src + byteParLineBW * ( sign>0 ? y :ly-1 -y );
            for( x=0; x<lx; x++ ) d[i++] = lut[getBW(here[x>>3],x)];
        }
    } else return 1;
    return 0;
}

int Check_bmpfile( void *s, unsigned long length ){
// メモリー上のデータが、処理可能な BMP ファイル形式であるかをチェックする
    if( s==NULL || length<=sizeof(HEADER) )  return 1;            // return 1;
    HEADER h=*(HEADER *)s;
    unsigned long isize=h.biSizeImage, fsize=isize+h.bfOffBits;
    unsigned long psize=h.biWidth*abs(h.biHeight)*(h.biBitCount/8);
    if( h.bfType!=BMPID || h.bfReserved1!=0 || h.bfReserved2!=0 )    return 2;
    else if( h.biPlanes!=1 || h.biCompression!=0 )                   return 3;
    else if( h.biSizeImage!=0 && (length<fsize || isize<psize) )     return 4;
    else if( h.biBitCount==32 && h.bfOffBits!=sizeof(HEADER) )       return 5;
    else if( h.biBitCount==24 && h.bfOffBits!=sizeof(HEADER) )       return 6;
    else if( h.biBitCount== 8 && (h.bfOffBits-sizeof(HEADER))%sizeof(PIXEL) ) {
       fprintf( stderr,"Unexpected lut size. %d, %d\n", h.bfOffBits,
              (int)((h.bfOffBits-sizeof(HEADER))/sizeof(PIXEL)) );   return 7;
    } else if(h.biBitCount==1 && (h.bfOffBits-sizeof(HEADER))!=2*sizeof(PIXEL)){
       fprintf( stderr,"Unexpected, requirement is 2, lut size. %d, %d\n",
              h.bfOffBits, (int)((h.bfOffBits-sizeof(HEADER))/sizeof(PIXEL)) );
                                                                     return 8;
    }
    return 0;
}

int file2memory( char *fname, char **image, long *size ){
// ディスク上のファイルを、メモリーに一括で読み込む
struct stat      fileInfo ;
FILE    *fp ;
char    *raw=NULL ;

    *image=NULL; *size=0;
    if ( (fp=fopen(fname,"rb")) == NULL ) {
        fprintf( stderr, "'%s' open error.\n", fname );               return 11;
    } else if ( stat(fname, &fileInfo) != 0 ) {
        fprintf( stderr, "'%s' stat error.\n", fname );  fclose(fp);  return 12;
    } else if( fileInfo.st_size <= (int)sizeof(HEADER) ) {
        fprintf( stderr, "'%s' file size erro. too small(%d Byte)\n",
                         fname, (int)fileInfo.st_size ); fclose(fp);  return 13;
    } else if( (raw=(char *)malloc(fileInfo.st_size)) == NULL ) {
        fprintf( stderr, "Couldn't allocate memory.\n" ) ; fclose(fp);return 14;
    } else if ( fread(raw, fileInfo.st_size, 1, fp) != 1 ) {
        fprintf( stderr, "'%s' read error.\n", fname );  
                                              free(raw); fclose(fp);  return 15;
    } else fclose(fp) ;
    *image= raw;
    *size = fileInfo.st_size;
    return 0;
}

void argumentAnalysis(int argc, char *argv[], char **iname, char **oname){
int  i, st=1;   // プログラム起動時の引数解析処理
    for( i=1; i<argc; i++ ) {
        if( argv[i][0]=='-' ) {
            switch(argv[i][1]) {
                case 'h': st=0;
                default : goto GOOD_BY;
            }
        }
        else if( *iname==NULL ) *iname = argv[i];
        else if( *oname==NULL ) *oname = argv[i];
        else goto GOOD_BY;
    }
    return;
GOOD_BY:;
    char *p=argv[0], *op="[-h] ";
    for( i=strlen(p); i>0 && p[i]!='\\' && p[i]!='/'; i-- );; if(i) i++;
    fprintf( stderr, "\nusage: %s %s[input_bmp [complex_out.dat]]\n\n", &p[i], op );
    fprintf( stderr, "\t-h: This option shows only this message.\n" );
    exit(st);
}

int main(int argc, char *argv[]){
long  st=1, filesize=0;
char  *raw=NULL, *ifname=NULL, *ofname=NULL;
PIXEL **source=NULL, **destination=NULL ;

    argumentAnalysis( argc, argv, &ifname, &ofname );    // 引数解析
    if( ifname==NULL ) ifname = IFNAMEDEFAULT;           // 入力ファイル名の設定
    if( ofname==NULL ) ofname = FNAMEIS(IMAGEPROCESSING);// 出力ファイル名の設定
    
    // 画像データの読み込み（ディスクからメモリーへ）
    if( (st=file2memory( ifname, &raw, &filesize ))!=0 ) {
        fprintf(stderr, "'%s' read error.\n", ifname );            goto GOOD_BY;
    }
    // 画像データの確認：（画像データの確認と、作業用環境の構築）
    if( (st=Check_bmpfile( raw, filesize )) ) {
        fprintf(stderr, "'%s' unknown file\n", ifname );           goto GOOD_BY;
    }
    
    HEADER *h=(HEADER *)raw;
    int    lx=h->biWidth, ly=abs(h->biHeight);
    // 作業用環境の構築：（画像データの確認と、作業用環境の構築）
    if (  (destination=(PIXEL**)alloc2Darray(sizeof(PIXEL),lx,ly)) == NULL ) {
        fprintf(stderr, "Couldn't allocate memory. destination\n");goto GOOD_BY;
    } else if( (source=(PIXEL**)alloc2Darray(sizeof(PIXEL),lx,ly)) == NULL ) {
        fprintf(stderr, "Couldn't allocate memory. source\n");     goto GOOD_BY;
    } //読み込んだ画像を処理バッファーに格納 PIXEL形式への変換を等含む
    bmp2pixel( raw, &source[0][0] ) ;

    // 実際の画像処理と、処理結果の保存, 画像処理ルーチンの呼び出し
    if( (st=IMAGEPROCESSING( source, destination, lx, ly ))==0 ) {
        saveArray2bmp32( ofname, destination, lx, ly );    // BMPファイルで保存
    } else if( st>0 ) fprintf( stderr, "Image processing is not completed.\n" );
    
GOOD_BY:; // 後始末とプログラムの終了
    free(source); free(destination); free(raw);
    return st;
}
